// ******************************************************************
//
// V1 ROM Helper - Encodes 18500Hz, Mono, PCM WAV files to MVS ADPCM
//
// ******************************************************************

#include "stdafx.h"
#include "SampleEncoding.h"


const int BUFFERSIZE = 1024;
const int MAGNITUDE = 1024;

static int step_size[49] = 
{
       16,  17,  19,  21,  23,  25,  28,  31,   34,   37,   41,   45,
       50,  55,  60,  66,  73,  80,  88,  97,  107,  118,  130,  143,
      157, 173, 190, 209, 230, 253, 279, 307,  337,  371,  408,  449,
      494, 544, 598, 658, 724, 796, 876, 963, 1060, 1166, 1282, 1411, 1552 
};

static int step_adj[16] = 
{
   -1, -1, -1, -1, 2, 5, 7, 9,
   -1, -1, -1, -1, 2, 5, 7, 9
};


// check file is correct PCM type
bool CSampleEncoding::ValidateFile(CString PathName, CString Filename)
{
   CFile fIn;

   if (!fIn.Open(PathName, CFile::modeRead))
   {
      CString ErrMsg;
      ErrMsg = "Failed to Open " + Filename;
      AfxMessageBox(ErrMsg);
      return false;
   }

   BYTE Value, Value2;

   // offset 0x16 = 1 for MONO
   fIn.Seek(0x16, CFile::begin);
   fIn.Read(&Value, 1);
   if (Value != 1)
   {
      AfxMessageBox("WAV file is not MONO");
      return false;
   }

   // offset 0x18 and 0x19 = 0x44,0x48 for 18500Hz
   fIn.Seek(0x18, CFile::begin);
   fIn.Read(&Value, 1);
   fIn.Read(&Value2, 1);
   if (Value != 0x44 && Value2 != 0x48)
   {
      AfxMessageBox("WAV file is not 18500Hz");
      return false;
   }

   // offset 0x20 = 2 for 2 bytes per sample
   fIn.Seek(0x20, CFile::begin);
   fIn.Read(&Value, 1);
   if (Value != 2)
   {
      AfxMessageBox("WAV file is not 16bit");
      return false;
   }

   // file is OK so save the details
   m_PathName = PathName;
   m_FileName = Filename;

   // store expected output size = 
   // input size - 44 bytes (for WAV RIFF header)
   // divided by 4 as ADPCM compresses a single 16bit sample to 4 bits
   m_WavLen = fIn.GetLength();
   m_ADPCMLen = (m_WavLen - 44) / 4;

   fIn.Close();
   return true;
}


int CSampleEncoding::GetModuleSize()
{
   return m_ADPCMLen;
}


// limit - a generic function that restricts a value between 2 limits
int CSampleEncoding::limit(int InVal, int MAX, int MIN)
{
   if (InVal > MAX) InVal = MAX;
   if (InVal < MIN) InVal = MIN;

   return InVal;
}

short CSampleEncoding::ADPCMDecode(BYTE code)
{
   short diff, e, ss;

   ss = step_size[m_Index];
   e = ss/8;

   if (code & 1) e += ss/4;
   if (code & 2) e += ss/2;
   if (code & 4) e += ss;
   if (code & 8) e *= -1;

   diff = e;

   m_Last = limit(m_Last + diff,MAGNITUDE-1,-MAGNITUDE);
   m_Index = limit(m_Index + step_adj[code], 48, 0);

   return m_Last;
}


BYTE CSampleEncoding::ADPCMEncode(short samp)
{
   short diff, ss;
   BYTE code;

   ss = step_size[m_Index];
   code = 0;

   diff = samp - m_Last;
   if (diff < 0) code = 8;

   diff = abs(diff);

   if (diff >= ss)
   {
      code |= 4;
      diff -= ss;
   }

   if (diff >= (ss/2))
   {
      code |= 2;
      diff -= (ss/2);
   }

   if (diff >= (ss/4))
   {
      code |= 1;
   }

   m_Last = ADPCMDecode(code);
   return code;
}


bool CSampleEncoding::Encode(CFile * fWrite)
{
   // open input WAV file - note it has already validated
   CFile fIn;
   if (!fIn.Open(m_PathName, CFile::modeRead))
      return false;
   
   // skip RIFF header
   fIn.Seek(44, CFile::begin);

   m_Index = 0;
   m_Last = 0;

   short Buffer12[BUFFERSIZE/2];

   while (1)
   {
      int i;

      int BytesRead = fIn.Read(Buffer12, BUFFERSIZE);
      if (BytesRead <= 0)
         break;
      
      // convert data down to 12 bits
      // BytesRead/2 as we are working in 16bit WORDs
      for (i = 0; i < BytesRead/2; i++)
      {
          short * pBuf = &Buffer12[i];
          *pBuf = *pBuf / 32;
      }

      // Generate ADPCM
      // BytesRead/4 as we are doing 2 WORDs at once because each
      // WORD is compressed to 4 bit.
      int outPos = 0;
      for (i = 0; i < BytesRead/2; i+=2)
      {
         BYTE Val;

         Val = ADPCMEncode(Buffer12[i]);
         Val <<= 4;
         Val |= ADPCMEncode(Buffer12[i+1]);

         fWrite->Write(&Val, 1);
      }
   }

   fIn.Close();
   return 0;
}
